package cn.tursom.web.netty

import cn.tursom.core.buffer.ByteBuffer
import cn.tursom.web.AdvanceHttpContent
import cn.tursom.web.utils.Chunked
import io.netty.buffer.ByteBuf
import io.netty.buffer.CompositeByteBuf
import io.netty.buffer.Unpooled
import io.netty.channel.ChannelHandlerContext
import io.netty.handler.codec.http.*
import io.netty.handler.stream.ChunkedFile
import java.io.File
import java.io.RandomAccessFile
import kotlin.collections.set

@Suppress("MemberVisibilityCanBePrivate", "unused")
open class NettyHttpContent(
  val ctx: ChannelHandlerContext,
  val msg: FullHttpRequest
) : AdvanceHttpContent, NettyResponseHeaderAdapter() {
  override var finished: Boolean = false
  override val uri: String by lazy {
    var uri = msg.uri()
    while (uri.contains("//")) {
      uri = uri.replace("//", "/")
    }
    uri
  }
  override val clientIp get() = ctx.channel().remoteAddress()!!
  override val realIp: String = super.realIp
  val httpMethod: HttpMethod get() = msg.method()
  val protocolVersion: HttpVersion get() = msg.protocolVersion()
  val headers: HttpHeaders get() = msg.headers()
  protected val paramMap by lazy { RequestParser.parse(msg) }
  override val cookieMap by lazy { getHeader("Cookie")?.let { decodeCookie(it) } ?: mapOf() }
  override val body = msg.content()?.let { NettyByteBuffer(it) }

  //override val responseBody = ByteArrayOutputStream()
  var responseStatus: HttpResponseStatus = HttpResponseStatus.OK
  override var responseCode: Int
    get() = responseStatus.code()
    set(value) {
      responseStatus = HttpResponseStatus.valueOf(value)
    }
  override var responseMessage: String? = null
  override val method: String get() = httpMethod.name()
  val chunkedList = ArrayList<ByteBuffer>()
  val responseBodyBuf: CompositeByteBuf = ctx.alloc().compositeBuffer()!!

  override fun getHeader(header: String): String? {
    return headers[header]
  }

  override fun getHeaders(header: String): List<String> {
    return headers.getAll(header)
  }

  override fun getHeaders(): Iterable<Map.Entry<String, String>> {
    return headers
  }

  override fun getParams(): Map<String, List<String>> {
    return paramMap
  }

  override fun getParams(param: String): List<String>? {
    return paramMap[param]
  }

  override fun addParam(key: String, value: String) {
    if (!paramMap.containsKey(key)) {
      paramMap[key] = ArrayList()
    }
    (paramMap[key] as ArrayList).add(value)
  }

  override fun write(message: String) {
    responseBodyBuf.addComponent(Unpooled.wrappedBuffer(message.toByteArray()))
    //responseBody.write(message.toByteArray())
  }

  override fun write(byte: Byte) {
    val buffer = ctx.alloc().buffer(1).writeByte(byte.toInt())
    responseBodyBuf.addComponent(buffer)
    //responseBody.write(byte.toInt())
  }

  override fun write(bytes: ByteArray, offset: Int, size: Int) {
    responseBodyBuf.addComponent(Unpooled.wrappedBuffer(bytes, offset, size))
    //responseBody.write(bytes, offset, size)
  }

  override fun write(buffer: ByteBuffer) {
    //buffer.writeTo(responseBody)
    responseBodyBuf.addComponent(if (buffer is NettyByteBuffer) {
      buffer.byteBuf
    } else {
      val buf = Unpooled.wrappedBuffer(buffer.readBuffer())
      buffer.clear()
      buf
    })
  }

  override fun reset() {
    responseBodyBuf.clear()
  }

  override fun finish() {
    finish(responseBodyBuf)
  }

  override fun finish(buffer: ByteArray, offset: Int, size: Int) {
    finish(Unpooled.wrappedBuffer(buffer, offset, size))
  }

  override fun finish(buffer: ByteBuffer) {
    if (buffer is NettyByteBuffer) {
      finish(buffer.byteBuf)
    } else {
      super.finish(buffer)
    }
  }

  fun finish(buf: ByteBuf) = finish(buf, responseStatus)
  fun finish(buf: ByteBuf, responseCode: HttpResponseStatus) {
    val response = DefaultFullHttpResponse(HttpVersion.HTTP_1_1, responseCode, buf)
    finish(response)
  }

  fun finish(response: FullHttpResponse) {
    finished = true
    val heads = response.headers()
    addHeaders(
      heads,
      mapOf(
        HttpHeaderNames.CONTENT_TYPE to "${HttpHeaderValues.TEXT_PLAIN}; charset=UTF-8",
        HttpHeaderNames.CONTENT_LENGTH to response.content().readableBytes(),
        HttpHeaderNames.CONNECTION to HttpHeaderValues.KEEP_ALIVE
      )
    )
    ctx.writeAndFlush(response)
  }

  override fun writeChunkedHeader() {
    val response = DefaultHttpResponse(HttpVersion.HTTP_1_1, HttpResponseStatus.OK)
    response.status = if (responseMessage != null) HttpResponseStatus(responseCode, responseMessage)
    else responseStatus
    val heads = response.headers()
    addHeaders(
      heads, mapOf(
      HttpHeaderNames.CONTENT_TYPE to "${HttpHeaderValues.TEXT_PLAIN}; charset=UTF-8",
      HttpHeaderNames.CONNECTION to HttpHeaderValues.KEEP_ALIVE,
      HttpHeaderNames.TRANSFER_ENCODING to "chunked"
    )
    )
    ctx.write(response)
  }

  override fun addChunked(buffer: ByteBuffer) {
    chunkedList.add(buffer)
  }

  override fun finishChunked() {
    finished = true
    val httpChunkWriter = HttpChunkedInput(NettyChunkedByteBuffer(chunkedList))
    ctx.writeAndFlush(httpChunkWriter)
  }

  override fun finishChunked(chunked: Chunked) {
    finished = true
    val httpChunkWriter = HttpChunkedInput(NettyChunkedInput(chunked))
    ctx.writeAndFlush(httpChunkWriter)
  }

  override fun finishFile(file: File, chunkSize: Int) {
    finished = true
    writeChunkedHeader()
    ctx.writeAndFlush(HttpChunkedInput(ChunkedFile(file, chunkSize)))
  }

  override fun finishFile(file: RandomAccessFile, offset: Long, length: Long, chunkSize: Int) {
    finished = true
    writeChunkedHeader()
    ctx.writeAndFlush(HttpChunkedInput(ChunkedFile(file, offset, length, chunkSize)))
  }
}

